/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef	AUDIO_HPP_
#define	AUDIO_HPP_

#include <compiler.hpp>
#include <strings.hpp>

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

struct audio_paths final {
    const char *mailbox{"none"};
    const char *app{"none"};
    const char *lang{"en"};
    const char *voice{"default"};
};

class audio_file final {
public:
    inline audio_file() = default;
    inline audio_file(const audio_file& from) = delete;

    inline ~audio_file() {
        close();
    }

    inline auto operator=(const audio_file& from) -> audio_file& = delete;

    inline operator bool() const {
        return stream_ != -1;
    }

    inline auto operator!() const {
        return stream_ == -1;
    }

    inline auto magic() const { // 0x2e736e64
        return get32_(&header_[0]);
    }

    inline auto offset() const { // at least 28....
        return get32_(&header_[4]);
    }

    inline auto size() const { // -1 = unknown
        return get32_(&header_[8]);
    }

    inline auto encoding() const { // 1=ulaw, 27=alaw, 3=16bit pcm
        return get32_(&header_[12]);
    }

    inline auto sample_rate() const { // 8000...
        return get32_(&header_[16]);
    }

    inline auto channels() const {  // 1...
        return get32_(&header_[20]);
    }

    inline auto annotation() const {
        return std::string_view(annotation_);
    }

    inline auto current() const {
        return static_cast<uint32_t>(lseek(stream_, 0L, SEEK_CUR));
    }

    inline auto read(void *buf, size_t size) const {    // FlawFinder: no
        return ::read(stream_, buf, size);  // FlawFinder: no
    }

    inline auto write(const void *buf, size_t size) const {
        return ::write(stream_, buf, size);
    }

    inline auto rewind() const {
        return static_cast<uint32_t>(lseek(stream_, offset(), SEEK_SET));
    }

    inline auto append() const {
        return static_cast<uint32_t>(lseek(stream_, 0L, SEEK_END));
    }

    void close() {
        if(stream_ != -1) {
            if(update_) {
                put32_(&header_[8], current() - offset());
                ::lseek(stream_, 0L, SEEK_SET);
                ::write(stream_, header_, sizeof(header_));
            }
            ::close(stream_);
            annotation_ = "";
            stream_ = -1;
            update_ = false;
        }
    }

    auto make(const char *filename, uint32_t type = 1, uint32_t rate = 8000, uint32_t count = 1, unsigned mode = 0666, std::string_view note = "") {
        close();
        auto note_size = note.size();
        if(note_size < 3)
            note_size = 4;
        else
            ++note_size;
        while((note_size % 4) != 0)
            ++note_size;
        put32_(&header_[0], 0x2e736e64);
        put32_(&header_[4], 24 + note_size);
        put32_(&header_[8], -1);
        put32_(&header_[12], type);
        put32_(&header_[16], rate);
        put32_(&header_[20], count);
        put32_(&header_[24], 0);
        stream_ = ::creat(filename, mode);
        if(stream_ < 0 || ::write(stream_, header_, 28) != 28) {
            close();
            return false;
        }
        lseek(stream_, 24, SEEK_SET);
        if(!note.empty())
            ::write(stream_, note.data(), note.size());
        rewind();
        update_ = true;
        return true;
    }

    auto open(const char *filename) { // FlawFinder: no
        close();
        stream_ = ::open(filename, O_RDONLY);   // FlawFinder: no
        if(stream_ < 0 || ::read(stream_, header_, 28) != 28) { // FlawFinder: no
fail:
            close();
            return false;
        }
        if(magic() != 0x2e736e64 || offset() > 256)
            goto fail;
        char buf[offset() - 23];
        if(::read(stream_, buf, offset() - 24) != offset() - 24) // FlawFinder: no
            goto fail;
        buf[sizeof(buf) - 1] = 0;
        annotation_ = buf;
        return true;
    }

    static auto path(const char *from, char *to, size_t size, const audio_paths *paths) -> bool
    {
        if(!from || !*from || !to || size < 2)
            return false;

        if(strstr(from, "..") || strstr(from, "/.") || *from == '.' || *from == '/')
            return false;

        if(eq(from, "greet/", 6)) {
            if(eq(paths->mailbox, "none"))
                return false;
            snprintf(to, size, "%s/greet/%s/%s%s", BORDEAUX_PROMPTS_PATH, paths->mailbox, from + 6, ".au");
        }
        else if(eq(from, "app/", 4))
            snprintf(to, size, "%s/app/%s/%s%s", BORDEAUX_VOICES_PATH, paths->app, from + 4, ".au");
        if(strchr(from, '/'))
            snprintf(to, size, "%s/%s%s", BORDEAUX_PROMPTS_PATH, from, ".au");
        else
            snprintf(to, size, "%s/%s/%s/%s%s", BORDEAUX_VOICES_PATH, paths->lang, paths->voice, from, ".au");
        while(to && *to) {
            *to = static_cast<char>(tolower(*to));
            ++to;
        }
        return true;
    }

private:
    static auto get32_(const uint8_t *ptr) -> uint32_t {
        return ((uint32_t)(ptr[0]) << 24) | ((uint32_t)(ptr[1]) << 16) |
            ((uint32_t)(ptr[2]) << 8) | (uint32_t)(ptr[3]);
    }

    static void put32_(uint8_t *ptr, uint32_t value) {
        ptr[0] = (value & 0xff000000) >> 24;
        ptr[1] = (value & 0x00ff0000) >> 16;
        ptr[2] = (value & 0x0000ff00) >> 8;
        ptr[3] = value & 0xff;
    }

    uint8_t header_[28]{};
    bool update_{false};
    int stream_{-1};
    std::string annotation_;
};

/*!
 * Generic stand-alone code to handle sun audio file format used by Bordeaux.
 * \file audio.hpp
 */
#endif
