/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef	CLOCK_HPP_
#define	CLOCK_HPP_

#include "compiler.hpp"

#include <chrono>
#include <cstdint>
#include <ctime>

namespace server {
using namespace std::chrono_literals;
using microseconds_t = uint64_t;
using milliseconds_t = uint32_t;
using system_time = std::chrono::system_clock::time_point;
using steady_time = std::chrono::steady_clock::time_point;
using hires_time = std::chrono::high_resolution_clock::time_point;
using millisecond_time = std::chrono::milliseconds;
using microsecond_time = std::chrono::microseconds;

// TODO: const std::int32_t inf = ~0U;

struct timer : public timespec {
public:
    timer(milliseconds_t msec = 0);

    auto operator=(milliseconds_t msec) -> timer&;	// timer set current + msec
    auto operator+=(milliseconds_t msec) -> timer&;	// timer increment
    auto operator-=(const timer& base) -> timer&;	// timer difference

    inline auto operator!() const {			// diff expired...
        return !tv_sec && !tv_nsec;
    }

    inline auto expired() const {
        return !tv_sec && !tv_nsec;
    }

    auto until() const -> milliseconds_t;			// typically for poll
};

inline auto operator==(const timer& lh, const timer& rh) {
    return lh.tv_sec == rh.tv_sec && lh.tv_nsec == rh.tv_nsec;
}

inline auto operator!=(const timer& lh, const timer& rh) {
    return lh.tv_sec != rh.tv_sec || lh.tv_nsec != rh.tv_nsec;
}

inline auto operator+(const timer& lh, milliseconds_t msec) {
    auto base = lh;
    base += msec;
    return base;
}

inline auto operator-(const timer& lh, const timer& rh) {
    auto base = lh;
    base -= rh;
    return base;
}

auto operator<(const timer& lh, const timer& rh) -> bool;
auto operator>(const timer& lh, const timer& rh) -> bool;
auto operator<=(const timer& lh, const timer& rh) -> bool;
auto operator>=(const timer& lh, const timer& rh) -> bool;
} // end namespace

namespace server::util {
inline auto current_time() {
    return std::chrono::system_clock::now();
}

inline auto current_time(time_t timestamp) {
    return std::chrono::system_clock::from_time_t(timestamp);
}

inline auto current_time(system_time& from) {
    return std::chrono::system_clock::to_time_t(from);
}

inline auto stable_clock() {
    return std::chrono::steady_clock::now();
}

inline auto local_time(const std::time_t& time) {
    std::tm local; // NOLINT
    localtime_r(&time, &local);
    return local;
}

inline auto steady_clock(milliseconds_t offset = 0) {
    return std::chrono::steady_clock::now() + std::chrono::milliseconds(offset);
}

inline auto hires_clock(microseconds_t offset = 0) {
    return std::chrono::high_resolution_clock::now() + std::chrono::microseconds(offset);
}

inline auto steady_duration(const steady_time& start, const steady_time& end) {
    return std::chrono::duration_cast<std::chrono::milliseconds>(end - start);
}

inline auto steady_elapsed(const steady_time& start) {
    return std::chrono::duration_cast<std::chrono::milliseconds>(steady_clock() - start);
}

inline auto steady_remains(const steady_time& end) {
    return std::chrono::duration_cast<std::chrono::milliseconds>(end - steady_clock());
}

inline auto hires_duration(const hires_time& start, const hires_time& end) {
    return std::chrono::duration_cast<std::chrono::microseconds>(end - start);
}

inline auto hires_elapsed(const hires_time& start) {
    return std::chrono::duration_cast<std::chrono::microseconds>(hires_clock() - start);
}

inline auto hires_remains(const hires_time& end) {
    return std::chrono::duration_cast<std::chrono::microseconds>(end - hires_clock());
}

void timespec_delay(timespec *current, milliseconds_t offset = 0);
void sleep(milliseconds_t msec);
void sleep(const steady_time& until);
auto sleep_until(const timer& until) -> bool;
auto to_duration(const char *cp) -> milliseconds_t;
} // end namespace

/*!
 * Clocking and timing functions.  This enhances chrono from c++14
 * and includes thread timing and process priority.\file clock.hpp
 */
#endif
