/*
 * Copyright (C) 2017-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MEDIA_HPP_
#define MEDIA_HPP_

#include "compiler.hpp"
#include "clock.hpp"
#include "network.hpp"
#include "utils.hpp"
#include "audio.hpp"

#include <thread>
#include <mutex>
#include <stdexcept>
#include <string>
#include <cstring>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/ip.h>

namespace server {
static const size_t rtp_min_header_size = 12;
static const size_t rtp_max_header_size = 12 + (15 * 4); // 15 csrc...
static const size_t max_audio_size = 160;

class Session;

class rtp_packet final {
public:
    explicit rtp_packet(uint8_t *p) :
    packet_(p), samples_(0) {
        memset(packet_, 0, rtp_min_header_size);
    }

    rtp_packet() = delete;
    rtp_packet(const rtp_packet&) = delete;

    inline auto version() const {  // default is 80/2
        return (packet_[0] & 0xc0) >> 6;
    }

    inline auto padding() const {
        return (packet_[0] & 0x20) != 0;
    }

    inline auto contrib() const {
        return (packet_[0] & 0x0f);
    }

    inline auto contrib4() const {
        return (packet_[0] & 0x0f) << 2;
    }

    inline auto marker() const {
        return (packet_[1] & 0x80) != 0;
    }

    inline auto payload() const {
        return (packet_[1] & 0x7f);
    }

    inline auto sequence() const {
        return util::get16(&packet_[2]);
    }

    inline auto timestamp() const {
        return util::get32(&packet_[4]);
    }

    inline auto source() const {
        return util::get32(&packet_[8]);
    }

    inline auto extends() const -> uint16_t {
        if(!(packet_[0] & 0x10))
            return 0;
        return util::get16(&packet_[rtp_min_header_size + contrib4() + 2]) + 4;
    }

    inline auto profile_id() const -> uint16_t {
        if(!extends())
            return 0;
        return util::get16(&packet_[rtp_min_header_size + contrib4()]);
    }

    inline auto profile_size() const -> uint16_t {
        if(!extends())
            return 0;
        return util::get16(&packet_[rtp_min_header_size + contrib4() + 2]);
    }

    inline auto header_size() const {
        return  rtp_min_header_size + contrib4() + extends();
    }

    inline auto framing() const {
        return framing_;
    }

    inline auto samples() const {
        return samples_;
    }

    inline auto profile_data() const {
        return (extends() ? &packet_[rtp_min_header_size + contrib4() + 4] : nullptr);
    }

    inline auto data() const {
        return packet_ + rtp_min_header_size + contrib4() + extends();
    }

    // type for packet, samples per frame, and millisecond frametime. I
    // initially am doing 20ms/8khz ulaw, but other possibilities may exist in
    // the future.
    inline void set_payload(uint8_t value, uint32_t spf = 160, milliseconds_t ft = 20) {
        value |= (packet_[1] & 0x80);
        packet_[1] = value;
        samples_ = spf;
        framing_ = ft;
    }

    inline void set_version(uint8_t value = 2) const {
        packet_[0] = (value << 6) | (packet_[0] & 0x3f);
    }

    inline void set_source(uint32_t value) {
        util::set32(&packet_[8], value);
    }

    // next frame...
    inline void inc_sequence(unsigned count = 1) const {
        util::set16(&packet_[2], sequence() + count);
        util::set32(&packet_[4], timestamp() + samples_ * count);
    }

    // skipped silent frame...
    inline void inc_silence(unsigned count = 1) const {
        util::set32(&packet_[4], timestamp() + samples_ * count);
    }

    inline void adjust_offset(uint32_t offset) {
        util::set32(&packet_[4], timestamp() - offset);
    }

    inline auto send(int socket, const struct sockaddr *addr, size_t max) const {
        auto size = ::sendto(socket, packet_, max + header_size(), MSG_DONTWAIT, addr, net::size(addr));
        if(size < 0)
            return size;
        if(size <= static_cast<ssize_t>(header_size()))
            return ssize_t(0);
        return ssize_t(size - header_size());
    }

    inline auto recv(int socket, struct sockaddr *addr, size_t max) const {
        socklen_t len{sizeof(struct sockaddr_storage)};
        auto size = ::recvfrom(socket, packet_, max + header_size(), 0, addr, &len);
        if(size < 0)
            return size;
        if(size <= static_cast<ssize_t>(header_size()))
            return ssize_t(0);
        return ssize_t(size - header_size());
    }

private:
    uint8_t *packet_;
    uint32_t samples_;
    milliseconds_t framing_{20};
};

class rtp_audio final {
public:
    uint16_t local_port{0}, remote_port{0};
    uint8_t payload{0}, dtmf{100};
    std::string media, remote_media;

    rtp_audio() = delete;
    rtp_audio(const rtp_audio&) = delete;
    ~rtp_audio();

    rtp_audio(Session *s, std::string local_media, std::string_view& sdp, uint16_t port = 0);

    inline operator bool() const {
        return active_;
    }

    inline auto operator!() const {
        return !active_;
    }

    inline void sync_with_origin(bool sync_flag) {
        sync_address = sync_flag;
    }

    inline auto address() {
        return reinterpret_cast<struct sockaddr *>(&remote_to);
    }

    inline auto from() {
        if(!sync_address)
            return reinterpret_cast<struct sockaddr *>(&remote_from);
        return address();
    }

    void apply_sdp(std::string_view& sdp);
    void start();
    void finish();
    void ignore();
    void keepalive();
    void record_file(const char *name);
    void play_file(const char *name);
    auto silent() -> unsigned;

private:
    std::thread thread_;        // exec thread
    std::mutex mutex_;          // lock gaurd
    Session *session_;          // session to post events to
    audio_file audio_;          // current audio file if any...
    int socket_{-1};
    struct sockaddr_storage remote_to, remote_from;
    bool sync_address{false};
    uint64_t sent_total, recv_total, fail_total, skip_total;
    uint32_t samples_{160};     // samples per frame...
    unsigned counted_{0};       // counted samples
    unsigned fps_{50};
    unsigned bytes_{1};
    uint8_t filler_{0xff};
    milliseconds_t framing_{20};
    bool active_{false};
    uint32_t offset_{0};
    std::chrono::steady_clock::time_point timer_;

    auto (rtp_audio::*get_)(rtp_packet& rtp) noexcept -> size_t;
    auto (rtp_audio::*put_)(rtp_packet& rtp) noexcept -> size_t;

    // audio packet buffers...
    uint8_t input_buffer[rtp_max_header_size + max_audio_size];
    uint8_t output_buffer[rtp_max_header_size + max_audio_size];
    rtp_packet input_packet, output_packet;

    // i/o processing...
    void exec_thread();
    auto sync_timer() noexcept -> milliseconds_t;
    auto output() noexcept -> size_t;
    auto input() noexcept -> size_t;

    // file/packet ops...
    auto put_silent(rtp_packet& rtp) noexcept -> size_t;
    auto put_ignore(rtp_packet& rtp) noexcept -> size_t;
    auto get_ignore(rtp_packet& rtp) noexcept -> size_t;
};
} // end namespace server

/*!
 * Basic media support and media related rtp processing.
 * \file media.hpp
 */
#endif
